#fitbuttonid{
    position: absolute;
    width: 3em;
    margin: 0.5em;
    top: 0;
    z-index: 9999;
    right: 6.8em;
    height: 1.8em;
}

#dbutton{
    position: absolute;
    width: 7em;
    margin: 0.5em;
    top: 0;
    z-index: 9999;
    right: 0;
    height: 2em;
}

.tt-menu {
    background: #fff;
    width: 11em;
    border: 1px solid #ccc;
    max-height: 150px;
    overflow-y: auto;
    margin-top: 0.5em;
}

.tt-suggestion:hover{
    background-color: yellow;
}

.tt-suggestion.tt-cursor {
    background-color: yellow;
}

.tt-menu * {
    margin: 0;
    padding: 0;
}

.tt-suggestion {
    margin: 0;
    padding: 0;
    cursor: pointer;
    color: black;
}

.tt-suggestion + .tt-suggestion {
    border-top: 1px solid #ccc;
}

#infoid{
    position: absolute;
    left: 100px;
    top: 3em;
    margin: 0.5em;
    background: #fff;
    width: 14em;
    border: 1px solid #ccc;
    box-shadow: inset 0 0 0 0.25em rgba(187, 219, 247, 0.9);
    display: none;
    overflow: auto;
}

#close {
    float:right;
    display:inline-block;
    padding:2px 5px;
    background:#ccc;
    cursor: pointer;
}

#close:hover {
    float:right;
    display:inline-block;
    padding:2px 5px;
    background:#ccc;
    cursor: pointer;
    color:#fff;
}

#title {
    position: absolute;
    left: 11.5em;
    top: 1.8em;
    width: 30%;
    z-index: 9999;
    overflow: auto;
}

#layoutList {
    position: absolute;
    width: 8em;
    top: 2.41em;
    z-index: 9999;
    right: 0.5em;
    height: 1.9em;
}

#search {
    width: 11em;
    font-family: inherit;
}

#searchDiv {
    position: absolute;
    left: 0;
    top: 1.5em;
    z-index: 9999;
    margin: 0.5em;
    width: 11em;
}

#pyviprid {
    position: absolute;
    left: 0;
    top: 0em;
    z-index: 9999;
    width: 11em;
}

#info {
    position: absolute;
    left: 0;
    top: 4em;
    margin: 0.5em;
    background: #fff;
    width: 14em;
    border: 1px solid #ccc;
    box-shadow: inset 0 0 0 0.25em rgba(187, 219, 247, 0.9);
    display: none;
    overflow: auto;
}

#myprocesses {
    position: absolute;
    width: 9em;
    margin: 0.5em;
    top: 1.9em;
    z-index: 9999;
    right: 11em;
    height: 1.8em;
}

#simLis {
    position: absolute;
    width: 3em;
    margin: 0.5em;
    top: 1.9em;
    z-index: 9999;
    right: 8em;
    height: 1.8em;
}

.loader {
    border: 16px solid #f3f3f3; /* Light grey */
    border-top: 16px solid #3498db; /* Blue */
    border-radius: 50%;
    width: 120px;
    height: 120px;
    animation: spin 2s linear infinite;
    position: absolute;
    left: 40%;
    top: 30%;
}

@keyframes spin {
    0% { transform: rotate(0deg); }
    100% { transform: rotate(360deg); }
}

.hide-loader{
    display:none;
}

.select-css {
	display: block;
	font-size: 13px;
	font-family: sans-serif;
	font-weight: 700;
	color: #444;
	line-height: 1.3;
	padding: .2em 1.5em .1em .3em;
	width: 100%;
	max-width: 100%;
	box-sizing: border-box;
	margin: 0;
	border: 1px solid #aaa;
	box-shadow: 0 1px 0 1px rgba(0,0,0,.04);
	border-radius: .5em;
	-moz-appearance: none;
	-webkit-appearance: none;
	appearance: none;
	background-color: #fff;
	background-image: url('data:image/svg+xml;charset=US-ASCII,%3Csvg%20xmlns%3D%22http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg%22%20width%3D%22292.4%22%20height%3D%22292.4%22%3E%3Cpath%20fill%3D%22%23007CB2%22%20d%3D%22M287%2069.4a17.6%2017.6%200%200%200-13-5.4H18.4c-5%200-9.3%201.8-12.9%205.4A17.6%2017.6%200%200%200%200%2082.2c0%205%201.8%209.3%205.4%2012.9l128%20127.9c3.6%203.6%207.8%205.4%2012.8%205.4s9.2-1.8%2012.8-5.4L287%2095c3.5-3.5%205.4-7.8%205.4-12.8%200-5-1.9-9.2-5.5-12.8z%22%2F%3E%3C%2Fsvg%3E'),
	  linear-gradient(to bottom, #ffffff 0%,#e5e5e5 100%);
	background-repeat: no-repeat, repeat;
	background-position: right .7em top 50%, 0 0;
	background-size: .65em auto, 100%;
}
.select-css::-ms-expand {
	display: none;
}
.select-css:hover {
	border-color: #888;
}
.select-css:focus {
	border-color: #aaa;
	box-shadow: 0 0 1px 3px rgba(59, 153, 252, .7);
	box-shadow: 0 0 0 3px -moz-mac-focusring;
	color: #222;
	outline: none;
}
.select-css option {
	font-weight:normal;
}