define((()=>(()=>{var r={758:(r,e,t)=>{t.p=document.querySelector("body").getAttribute("data-base-url")+"nbextensions/pyvipr",window.require&&window.require.config({map:{"*":{pyvipr:"nbextensions/pyvipr/index"}}}),r.exports={load_ipython_extension:function(){}}}},e={};function t(n){var i=e[n];if(void 0!==i)return i.exports;var o=e[n]={exports:{}};return r[n](o,o.exports,t),o.exports}return t.g=function(){if("object"==typeof globalThis)return globalThis;try{return this||new Function("return this")()}catch(r){if("object"==typeof window)return window}}(),(()=>{var r;t.g.importScripts&&(r=t.g.location+"");var e=t.g.document;if(!r&&e&&(e.currentScript&&(r=e.currentScript.src),!r)){var n=e.getElementsByTagName("script");n.length&&(r=n[n.length-1].src)}if(!r)throw new Error("Automatic publicPath is not supported in this browser");r=r.replace(/#.*$/,"").replace(/\?.*$/,"").replace(/\/[^\/]+$/,"/"),t.p=r})(),t(758)})()));